{-|
Module      : Alvis
Description : General Alvis Compiler functions
Copyright   : (c) 2011-2015 by Marcin Szpyrka, Michal Wypych
License     : GPL-3
Maintainer  : mszpyrka@agh.edu.pl
Stability   : experimental

The Alvis module contains general function used by the
Alvis compiler.
-}

module Alvis where


-- | The set of Alvis agents' modes.
data Mode 
  = F     -- ^ finished
  | I     -- ^ init
  | R     -- ^ ready
  | T     -- ^ taken
  | W     -- ^ waiting
  | X     -- ^ running (executing)
  deriving (Eq,Show)

----------------------------------------------------------------------

-- | The set of contents information list entries.
data ContentsInfo 
  = CProc  String    -- ^ Name of called procedure.
  | CIn    String    -- ^ Name of accessible in procedure or name of port with executed in statement.
  | COut   String    -- ^ Name of accessible out procedure or name of port with executed out statement.
  | CTimer Int       -- ^ Number of timer time-units to go by.
  | CSFT   Int       -- ^ Number of time-units to the step finish time.
  deriving (Eq)

----------------------------------------------------------------------

-- | The 'ContentsInfo' type show function.
instance Show ContentsInfo where
  show (CProc a)  = "proc(" ++ a ++ ")"
  show (CIn a)    = "in(" ++ a ++ ")"
  show (COut a)   = "out(" ++ a ++ ")"
  show (CTimer n) = "timer(" ++ (show n) ++ ")"
  show (CSFT n)   = "sft(" ++ (show n) ++ ")"

----------------------------------------------------------------------

{-|
  The set of Alvis transitions from the Haskell 
  representation point of view. Some of the transitions e.g.
  'TIn', 'TInAP', 'TInPP', 'STInAP', 'STInPP' are represented
  by the same label in the LTS graph.
-}
data TTransition 
  = TExec    String Int          -- ^ exec statement, /parameters:/ name of the agent, number of the statement.
  | TExit    String Int          -- ^ exit statement, /parameters:/ name of the agent, number of the statement.
  | TIf      String Int          -- ^ if statement, /parameters:/ name of the agent, number of the statement.
  | TIn      String Int          -- ^ in statement, /parameters:/ port of the agent (Agent.port), number of the statement.
  | TInAP    String String Int   -- ^ in statement - active agent calls (in statement) an output procedure, /parameters:/ port of the first agent (Agent.port), port of the second agent,  number of the statement.
  | TInPP    String String Int   -- ^ in statement - passive agent calls (in statement) an output procedure, /parameters:/ port of the first agent (Agent.port), port of the second agent,  number of the statement.
  | TInF     String String Int   -- ^ in statement - active agent finishes a communication with another active agent, /parameters:/ port of the first agent (Agent.port), port of the second agent,  number of the statement.
  | TJump    String Int          -- ^ jump statement, /parameters:/ name of the agent, number of the statement.
  | TLoop    String Int          -- ^ loop statement, /parameters:/ name of the agent, number of the statement.
  | TNull    String Int          -- ^ null statement, /parameters:/ name of the agent, number of the statement.
  | TOut     String Int          -- ^ out statement - active agent initialises a communication /parameters:/ port of the agent (Agent.port), number of the statement.
  | TOutAP   String String Int   -- ^ out statement - active agent calls (out statement) an input procedure, /parameters:/ port of the first agent (Agent.port), port of the second agent,  number of the statement.
  | TOutPP   String String Int   -- ^ out statement - passive agent calls (out statement) an input procedure, /parameters:/ port of the first agent (Agent.port), port of the second agent,  number of the statement.
  | TOutF    String String Int   -- ^ out statement - active agent finishes a communication with another active agent, /parameters:/ port of the first agent (Agent.port), port of the second agent,  number of the statement.
  | TSelect  String Int          -- ^ select statement, /parameters:/ name of agent, number of the statement.
  | TStart   String Int          -- ^ start statement, /parameters:/ name of the agent, number of the statement.
  | STInAP   String String Int   -- ^ System version of the 'TInAP' constructor - for wake up purposes.
  | STInPP   String String Int   -- ^ System version of the 'TInPP' constructor - for wake up purposes.
  | STOutAP  String String Int   -- ^ System version of the 'TOutAP' constructor - for wake up purposes.
  | STOutPP  String String Int   -- ^ System version of the 'TOutPP' constructor - for wake up purposes.
  deriving (Eq)               

----------------------------------------------------------------------

-- | The 'TTransition' type show function.
instance Show TTransition where
  show (TExec a _)     = "exec(" ++ a ++ ")"
  show (TExit a _)     = "exit(" ++ a ++ ")"
  show (TIf a _)       = "if(" ++ a ++ ")"
  show (TIn a _)       = "in(" ++ a ++ ")"
  show (TInAP a _ _)   = "in(" ++ a ++ ")"
  show (TInPP a _ _)   = "in(" ++ a ++ ")"
  show (TInF a _ _)    = "in(" ++ a ++ ")"
  show (TJump a _)     = "jump(" ++ a ++ ")"
  show (TLoop a _)     = "loop(" ++ a ++ ")"
  show (TNull a _)     = "null(" ++ a ++ ")"
  show (TOut a _)      = "out(" ++ a ++ ")"
  show (TOutAP a _ _)  = "out(" ++ a ++ ")"
  show (TOutPP a _ _)  = "out(" ++ a ++ ")"
  show (TOutF a _ _)   = "out(" ++ a ++ ")"
  show (TSelect a _)   = "select(" ++ a ++ ")"
  show (TStart a _)    = "start(" ++ a ++ ")"
  show (STInAP a _ _)  = "in(" ++ a ++ ")"
  show (STInPP a _ _)  = "in(" ++ a ++ ")"
  show (STOutAP a _ _) = "out(" ++ a ++ ")"
  show (STOutPP a _ _) = "out(" ++ a ++ ")"

----------------------------------------------------------------------

-- | Removes all occurrences of the element from the list.
remove :: (Eq a) => a    -- ^ element to be removed
                 -> [a]  -- ^ input list
                 -> [a]  -- ^ output list
remove y [] = []
remove y (x:xs) | y == x    = remove y xs
                | otherwise = x : remove y xs   

----------------------------------------------------------------------

-- | Returns True if the first list is prefix of the second list,
-- and returns False otherwise
prefix :: Ord a => [a]   -- ^ first list
                -> [a]   -- ^ second list
                -> Bool  -- ^ return value
prefix [] ys = True
prefix (x:xs) [] = False
prefix (x:xs) (y:ys) = (x == y) && prefix xs ys

----------------------------------------------------------------------

-- | Search a context information list for a 'CProc' entry, which
-- parameter is equal to one the agent (first parameter) procedure.
-- Returns /CProc ""/, if such an entry does not exist.
findproc :: String           -- ^ name of an agent
         -> [ContentsInfo]   -- ^ context information list
         -> ContentsInfo     -- ^ return 'CProc' entry
findproc _ [] = CProc ""
findproc agent ((CProc port):xs) 
  | (prefix agent port) = CProc port
  | otherwise = findproc agent xs
findproc agent (_:xs) = findproc agent xs

----------------------------------------------------------------------

-- | Returns True if the context list does not contain a 'CProc' entry,
-- and returns False otherwise
procfree :: [ContentsInfo]    -- ^ context information list
         -> Bool              -- ^ return value
procfree [] = True
procfree ((CProc _):_) = False
procfree (_:xs) = procfree xs

----------------------------------------------------------------------

-- | Returns the first port of the communication transition.
fstPort :: TTransition    -- ^ communication transition
        -> String         -- ^ name of the port (Agent.port)
fstPort (TInAP   port _ _) = port
fstPort (TInPP   port _ _) = port
fstPort (STInAP  port _ _) = port
fstPort (STInPP  port _ _) = port
fstPort (TInF    port _ _) = port
fstPort (TOutAP  port _ _) = port 
fstPort (TOutPP  port _ _) = port
fstPort (STOutAP port _ _) = port
fstPort (STOutPP port _ _) = port
fstPort (TOutF   port _ _) = port

----------------------------------------------------------------------

-- | Returns the second port of the communication transition.
sndPort :: TTransition    -- ^ communication transition
        -> String         -- ^ name of the port (Agent.port)
sndPort (TInAP   _ port _) = port
sndPort (TInPP   _ port _) = port
sndPort (TInF    _ port _) = port
sndPort (TOutAP  _ port _) = port 
sndPort (TOutPP  _ port _) = port
sndPort (TOutF   _ port _) = port
sndPort (STInAP  _ port _) = port
sndPort (STInPP  _ port _) = port
sndPort (STOutAP _ port _) = port
sndPort (STOutPP _ port _) = port

----------------------------------------------------------------------

-- | Returns True if the transition is a communication transition,
-- and returns False otherwise.
isInOutTransition :: TTransition    -- ^ transition
                  -> Bool           -- ^ return value
isInOutTransition (TInAP _ _ _)   = True
isInOutTransition (TInPP _ _ _)   = True
isInOutTransition (STInAP _ _ _)  = True
isInOutTransition (STInPP _ _ _)  = True
isInOutTransition (TInF _ _ _)    = True
isInOutTransition (TOutAP _ _ _)  = True
isInOutTransition (TOutPP _ _ _)  = True
isInOutTransition (STOutAP _ _ _) = True
isInOutTransition (STOutPP _ _ _) = True
isInOutTransition (TOutF _ _ _)   = True
isInOutTransition _               = False

----------------------------------------------------------------------

-- | Returns name of the agent for the given port.
agentName :: String    -- ^ port name (Agent.port)
          -> String    -- ^ return port name
agentName s = takeWhile ('.'/=) s   

----------------------------------------------------------------------

-- | Removes from the input list all communication transitions that 
-- use the given port as the second argument.
removeTransitions :: String           -- ^ port name (Agent.port)
                  -> [TTransition]    -- ^ input list of transitions
                  -> [TTransition]    -- ^ output list of transitions
removeTransitions _ [] = []
removeTransitions port (tr:trs) | isInOutTransition tr && port == sndPort tr = removeTransitions port trs
                                | otherwise = tr : removeTransitions port trs

----------------------------------------------------------------------

-- | Returns value of the lowest priority. The highest priority in Alvis is equal to 0.
-- The lowest priority in Alvis is equal to 9.
minPriority :: Int
minPriority = 9

----------------------------------------------------------------------
----------------------------------------------------------------------
----------------------------------------------------------------------
----------------------------------------------------------------------


